# -*- coding: utf-8 -*-

"""This file contains the ZucchiniState class that provides context for the
grader. The context should include the grader's local configuration,
the Assignment object related to the assignment in the present directory if
one exists, and an instance of the FarmManager class."""

import os

import yaml

from .canvas import CanvasAPI
from .amazon import AmazonAPI
from .utils import ConfigDictMixin
from .assignment import Assignment
from .farms import FarmManager
from .constants import FARM_DIRECTORY


class ZucchiniConfig(ConfigDictMixin):
    """Holds user configuration from user.yml"""

    def __init__(self, config_directory, user_name, user_email, canvas_url='',
                 canvas_token='', aws_access_key_id='',
                 aws_secret_access_key='', aws_s3_bucket_name=''):
        self.config_directory = config_directory

        self.user_name = user_name
        self.user_email = user_email

        self.canvas_url = canvas_url
        self.canvas_token = canvas_token

        self.farm_manager = FarmManager(
            os.path.join(config_directory, FARM_DIRECTORY))

        self.aws_access_key_id = aws_access_key_id
        self.aws_secret_access_key = aws_secret_access_key
        self.aws_s3_bucket_name = aws_s3_bucket_name


class ZucchiniState(object):
    def __init__(self, assignment_directory):
        self.assignment_directory = assignment_directory
        self.submission_dir = None
        self._assignment = None
        self._config = None

    def get_assignment(self):
        """
        Return an Assignment instance based on the configuration for this
        assignment.
        """

        if self._assignment is None:
            self._assignment = Assignment(self.assignment_directory)

        return self._assignment

    def canvas_api(self):
        """
        Return a CanvasAPI instance configured according to the user's global
        configuration.
        """

        if not self._config or not self._config.canvas_url or \
                not self._config.canvas_token:
            raise ValueError('The Canvas API is not configured!')

        return CanvasAPI(self._config.canvas_url, self._config.canvas_token)

    def get_amazon_api(self):
        """
        Return an AmazonAPI instance configured according to the user's global
        configuration.
        """

        if not self._config or \
                not self._config.aws_access_key_id or \
                not self._config.aws_secret_access_key or \
                not self._config.aws_s3_bucket_name:
            raise ValueError('The Amazon API is not configured')

        return AmazonAPI(
            self._config.aws_access_key_id,
            self._config.aws_secret_access_key,
            self._config.aws_s3_bucket_name)

    @property
    def user_name(self):
        return self._config.user_name

    @property
    def farm_manager(self):
        return self._config.farm_manager

    def save_config_to_file(self, config_file):
        config_dict = self._config.to_config_dict('config_directory')
        yaml.safe_dump(config_dict, config_file, default_flow_style=False)

    def load_config_from_dict(self, config_dict, config_directory):
        """Load user configuation from a dictionary."""
        self._config = ZucchiniConfig.from_config_dict(
            config_dict, config_directory=config_directory)

    def load_config_from_file(self, config_file, config_directory):
        """Load user configuration from a file-like object as yaml."""
        config_dict = yaml.safe_load(config_file)
        self.load_config_from_dict(config_dict, config_directory)



import re
from fractions import Fraction
import requests
import json


from . import PenalizerInterface, InvalidPenalizerConfigError

"""Penalize a submission after being checked off."""


class CheckoffPenalizer(PenalizerInterface):
    UNITS_REGEX = re.compile(r'^(?P<mag>[0-9/]+)\s*(?P<unit>[a-z]*)$')
    """
    Penalize students for submitting after being checked off
    on the attendance sheet.

    Configure it like this in the assignment config file. In this
    example, after 8 hours late you get 75 points off your grade (a 85
    would go to a 10).

    penalties:
    - name: CHECKOFF
      backend: CheckoffPenalizer
      backend-options:
        penalty: 100pts
        api-url: https:/.....
        api-key: djfal;ksf

    """

    def __init__(self, penalty, api_url, api_key):
        self.penalty, penalty_unit = self.split_units(penalty)
        if penalty_unit in ('pt', 'pts'):
            self.penalty /= 100
            self.penalty_points = True
        elif penalty_unit is None:
            self.penalty_points = False
        else:
            raise InvalidPenalizerConfigError("unknown penalty unit `{}'. try "
                                              "a fraction optionally followed "
                                              "by `pt'."
                                              .format(penalty_unit))

        self.api_url = api_url
        self.api_key = api_key

    def adjust_grade(self, submission, grade):
        params = {
            "student_name": submission.student_name,
            "attendance_event": submission.assignment.name
        }
        check = requests.post(url=self.api_url, data=json.dumps(params),
                              headers={"x-api-key": self.api_key})
        if check.json():
            if self.penalty_points:
                grade = max(0, grade - self.penalty)
                return grade
            else:
                grade = grade * (1 - self.penalty)
                return grade
        return grade

    @classmethod
    def split_units(cls, amount_str):
        if isinstance(amount_str, (int, float)):
            return Fraction(amount_str), None

        match = cls.UNITS_REGEX.match(amount_str.lower())

        if match is None:
            raise InvalidPenalizerConfigError("unknown units format `{}'"
                                              .format(amount_str))

        return Fraction(match.group('mag')), match.group('unit') or None



import json
from fractions import Fraction

from ..submission import BrokenSubmissionError
from ..utils import run_process, PIPE, TimeoutExpired
from ..grades import PartGrade
from . import GraderInterface, Part

"""
Grade a homework using a CircuitSim grader
"""


class CircuitSimTest(Part):
    __slots__ = ('test',)

    def __init__(self, test):
        self.test = test

    def description(self):
        return self.test

    def grade(self, result):
        if result is None:
            msg = ('Results for test not found. Check if there were any'
                   ' internal errors reported. If not, report this as an'
                   ' autograder error to your instructors.')
            return PartGrade(score=Fraction(0), log=msg)

        log = '\n'.join('{}: {}'.format(
            failure['displayName'],
            failure.get('message', '(no details, sorry)'))
            for failure in result['partialFailures'])
        partialFailures = len(result['partialFailures'])
        if partialFailures < result['failed']:
            log += '\n[omitted {} more failures]' \
                   .format(result['failed'] - partialFailures)

        score = Fraction(result['total'] - result['failed'], result['total'])
        return PartGrade(score=score, log=log)


class CircuitSimGrader(GraderInterface):
    """
    Run a CircuitSim grader based on
    <https://github.com/ausbin/circuitsim-grader-template> and collect
    the results.
    """

    DEFAULT_TIMEOUT = 30

    def __init__(self, grader_jar, test_class, timeout=None):
        self.grader_jar = grader_jar
        self.test_class = test_class

        if timeout is None:
            self.timeout = self.DEFAULT_TIMEOUT
        else:
            self.timeout = timeout

    def list_prerequisites(self):
        # CircuitSim needs JavaFX
        return ['openjdk-8-jre', 'openjfx']

    def needs_display(self):
        # CircuitSim needs JavaFX which needs a display server
        return True

    def part_from_config_dict(self, config_dict):
        return CircuitSimTest.from_config_dict(config_dict)

    def grade(self, submission, path, parts):
        cmdline = ['java', '-jar', self.grader_jar, '--zucchini',
                   self.test_class]
        try:
            # Do not mix stderr into stdout because sometimes our friend
            # Roi printStackTrace()s or System.err.println()s, and that
            # will mess up JSON parsing
            process = run_process(cmdline, cwd=path, timeout=self.timeout,
                                  stdout=PIPE, stderr=PIPE, input='')
        except TimeoutExpired:
            raise BrokenSubmissionError('timeout of {} seconds expired for '
                                        'grader'.format(self.timeout))

        if process.returncode != 0:
            raise BrokenSubmissionError(
                'grader command exited with nonzero exit code {}'
                .format(process.returncode),
                verbose='\n'.join(output_stream.decode()
                                  for output_stream
                                  in [process.stderr, process.stdout]
                                  if output_stream))

        results = json.loads(process.stdout.decode())

        if 'error' in results:
            raise BrokenSubmissionError(results['error'])

        method_results = {result['methodName']:
                          result for result in results['tests']}
        return [part.grade(method_results.get(part.test)) for part in parts]



import sys
import errno
import os
import re
import inspect
import glob
import shutil
import threading
from collections import namedtuple

import arrow
import click

if sys.version_info[0] < 3:
    # subprocess32, the python 3 subprocess module backported to python
    # 2, does not support non-POSIX platforms (aka Windows). So if
    # you're on Windows, you need to use python 3, since it has a more
    # up-to-date subprocess module in the standard library
    if os.name != 'posix':
        raise NotImplementedError('You need to use Python 3 on non-POSIX '
                                  'platforms')

    # Python 2 imports
    import Queue as queue  # noqa
    import subprocess32 as subprocess
    from urlparse import urlparse
else:
    # Python 3 imports
    import queue  # noqa
    import subprocess
    from urllib.parse import urlparse


def mkdir_p(path):
    try:
        os.makedirs(path)
    except OSError as exc:  # Python >2.5
        if exc.errno == errno.EEXIST and os.path.isdir(path):
            pass
        else:
            raise


def recursive_get_using_string(collection, key):
    """
    Given a collection and a key in the format of x.y.z.a, return collection
    [x][y][z][a].
    """

    if "." not in key:
        if key.isdigit():
            key = int(key)

        return collection[key]

    left, right = key.split('.', 1)
    return recursive_get_using_string(
            recursive_get_using_string(collection, left),
            right)


def run_thread(func, args, result_queue):
    """
    Run a thread which runs func(args), putting each yielded result in
    queue. Return the thread.
    """

    def thread():
        try:
            for result in func(*args):
                result_queue.put(result)
        except Exception as err:
            result_queue.put(err)

    thread = threading.Thread(target=thread)
    thread.start()
    return thread


# Patch around old versions of subprocess
PIPE = subprocess.PIPE
STDOUT = subprocess.STDOUT
TimeoutExpired = subprocess.TimeoutExpired
CompletedProcess = namedtuple('CompletedProcess', ('args', 'returncode',
                                                   'stdout', 'stderr'))


def run_process(*popenargs, **kwargs):
    """
    A straight copy-paste of subprocess.run() from the CPython source to
    support Python versions earlier than 3.5.
    """

    # Can't put these directly in function signature because PEP-3102 is
    # not a thing in Python 2
    input = kwargs.pop('input', None)
    timeout = kwargs.pop('timeout', None)
    check = kwargs.pop('check', False)

    if input is not None:
        if 'stdin' in kwargs:
            raise ValueError('stdin and input arguments may not both be used.')
        kwargs['stdin'] = subprocess.PIPE

    with subprocess.Popen(*popenargs, **kwargs) as process:
        try:
            stdout, stderr = process.communicate(input, timeout=timeout)
        except subprocess.TimeoutExpired:
            process.kill()
            stdout, stderr = process.communicate()
            raise subprocess.TimeoutExpired(process.args, timeout,
                                            output=stdout, stderr=stderr)
        except: # noqa
            process.kill()
            process.wait()
            raise
        retcode = process.poll()
        if check and retcode:
            raise subprocess.CalledProcessError(retcode, process.args,
                                                output=stdout, stderr=stderr)

    return CompletedProcess(process.args, retcode, stdout, stderr)


def sanitize_path(path, path_lib=os.path, join=True):
    """
    Convert an untrusted path to a relative path.

    Defaults to using os.path to manipulate (native) paths, but you can
    pass path_lib=posixpath to always use Unix paths, for example.

    If `join=False', return a list of path components. Default (True) is
    to path_lib.join() them.
    """

    # Remove intermediate ..s
    path = path_lib.normpath(path)
    # Remove leading /s
    path = path.lstrip(path_lib.sep)
    components = path.split(path_lib.sep)

    # Remove leading ..s and DOS drive letters
    while components and (components[0] == '..' or
                          len(components[0]) == 2 and components[0][1] == ':'):
        components = components[1:]

    if join:
        return path_lib.join(*components)
    else:
        return components


# Support FileNotFoundError, which does not exist in Python 2
try:
    FileNotFoundError = FileNotFoundError
except NameError:
    class FileNotFoundError(Exception):
        pass


def copy_globs(globs, src_dir, dest_dir):
    """
    Copy files matched by `globs' (a list of glob strings) from src_dir
    to dest_dir, maintaining directories if possible.
    """

    files_to_copy = []

    # Do a first pass to check for missing files. This way, we don't
    # copy a bunch of files only to blow up when we can't find a
    # later file.
    for file_glob in globs:
        absolute_glob = os.path.join(src_dir, file_glob)
        matches = glob.iglob(absolute_glob)

        if not matches:
            raise FileNotFoundError("missing file `{}'".format(file_glob))

        files_to_copy += matches

    for file_to_copy in files_to_copy:
        relative_path = os.path.relpath(file_to_copy, start=src_dir)
        dirname = os.path.dirname(relative_path)

        dest = os.path.join(dest_dir, relative_path)
        if os.path.isdir(file_to_copy):
            shutil.copytree(file_to_copy, dest)
        else:
            mkdir_p(os.path.join(dest_dir, dirname))
            shutil.copy(file_to_copy, dest)


# Same as the Canvas date format
_DATETIME_FORMAT = '%Y-%m-%dT%H:%M:%SZ'


def datetime_from_string(date_str):
    """
    Convert a human-readable date/time string (in the format used by
    Canvas and in submission metadata files) to a datetime instance.
    """

    return arrow.get(date_str).to('utc').datetime


def datetime_to_string(datetime_obj):
    """Convert a datetime UTC instance to a human-readable date/time string."""

    return datetime_obj.strftime(_DATETIME_FORMAT)


def current_iso8601():
    return arrow.now().replace(microsecond=0).isoformat()


class ConfigDictMixin(object):
    @staticmethod
    def _to_field(config_key):
        """Convert a config key to a field name"""
        return config_key.replace('-', '_')

    @staticmethod
    def _to_config(field_name):
        """Convert a field name to a config key"""
        return field_name.replace('_', '-')

    @classmethod
    def _find_args(cls, exclude_args):
        """
        Find the required and optional arguments of the constructor for
        the class `cls'. Returns a tuple containing two lists: the first
        contains required args; the second contains optional args.

        Exclude arguments found in `exclude_args'.
        """

        arg_spec = inspect.getfullargspec(cls.__init__)
        num_optional_args = 0 if arg_spec.defaults is None \
            else min(len(arg_spec.args)-1, len(arg_spec.defaults))
        first_optional_arg = len(arg_spec.args) - num_optional_args

        required_args = [arg for arg in arg_spec.args[1:first_optional_arg]
                         if arg not in exclude_args]
        optional_args = [arg for arg in arg_spec.args[first_optional_arg:]
                         if arg not in exclude_args]

        return required_args, optional_args

    def to_config_dict(self, *exclude):
        """
        Try to convert an instance to a configuration dictionary by
        accessing fields named after constructor fields.
        """

        required_args, optional_args = self._find_args(exclude)
        result = {}

        for arg in required_args:
            result[self._to_config(arg)] = getattr(self, arg)

        for arg in optional_args:
            if hasattr(self, arg) and getattr(self, arg) is not None:
                result[self._to_config(arg)] = getattr(self, arg)

        return result

    @classmethod
    def from_config_dict(cls, config_dict, **extra_kwargs):
        """
        Convert a dictionary, checking for invalid keys, that can be
        safely passed as kwargs. Checks only for the presence of options
        (keys), not their types. Converts dashes to underscores in
        option names.

        Considers optional arguments to cls.__init__() optional options,
        and required arguments to cls.__init__() required options.
        Rejects options also found in extra_kwargs.
        """

        # If already deserialized, just return
        if isinstance(config_dict, cls):
            return config_dict

        required_args, optional_args = cls._find_args(extra_kwargs)

        kwargs = {}

        for raw_key in config_dict:
            key = cls._to_field(raw_key)

            if key not in required_args + optional_args:
                raise ValueError("Unknown config key `{}'".format(raw_key))

            kwargs[key] = config_dict[raw_key]

        for key in required_args:
            if key not in kwargs:
                raw_key = cls._to_config(key)
                raise ValueError("Missing required config key `{}'"
                                 .format(raw_key))

        kwargs.update(extra_kwargs)
        return cls(**kwargs)


class ConfigDictNoMangleMixin(object):
    """
    A mixin which does not mangle config keys by converting `-' in
    config keys to `_' in field names (and vice versa) as
    ConfigDictMixin does by default.
    """

    @staticmethod
    def _to_field(config_key):
        """Convert a config key to a field name"""
        return config_key

    @staticmethod
    def _to_config(field_name):
        """Convert a field name to a config key"""
        return field_name


class Record(object):
    """
    A struct in Python, basically.
    Constructor sets fields according to the fields in __slots__.
    """

    def __init__(self, **kwargs):
        for arg, val in kwargs.items():
            if arg in self.__slots__:
                setattr(self, arg, val)
            else:
                raise TypeError('invalid argument to {}: {}'
                                .format(type(self), arg))

    def __repr__(self):
        props = ', '.join('{}={}'.format(prop, str(getattr(self, prop)))
                          for prop in self.__slots__ if hasattr(self, prop))
        return '<{} {}>'.format(type(self).__name__, props)


class CanvasURLType(click.ParamType):
    name = 'http(s) url'

    def convert(self, value, param, ctx):
        parsed = urlparse(value)
        # Be picky and demand https://... and no querystrings or
        # fragments or other weird stuff
        valid_https_url = parsed.scheme == 'https' \
            and '.' in parsed.netloc and parsed.path in ('/', '') \
            and not parsed.params and not parsed.query \
            and not parsed.fragment and not parsed.username \
            and not parsed.password

        if not valid_https_url:
            self.fail("`{}' is not a valid Canvas URL. I want something "
                      "like https://gatech.instructure.com/"
                      .format(value), param, ctx)

        # Strip trailing slash
        if value.endswith('/'):
            value = value[:-1]

        return value


CANVAS_URL = CanvasURLType()


class AwsAccessKeyIdType(click.ParamType):
    name = 'AWS Access Key ID'

    def convert(self, value, param, ctx):
        # TODO: actual error handling for AWS Access Key ID
        return value.strip()


AWS_ACCESS_KEY_ID = AwsAccessKeyIdType()


class AwsSecretAccessKeyType(click.ParamType):
    name = 'AWS Secret Access Key'

    def convert(self, value, param, ctx):
        # TODO: actual error handling for AWS Secret Access Key
        return value.strip()


AWS_SECRET_ACCESS_KEY = AwsSecretAccessKeyType()


class AwsBucketNameType(click.ParamType):
    name = 'AWS Bucket Name'

    def convert(self, value, param, ctx):
        # TODO: actual error handling for AWS Bucket Name
        return value.strip()


AWS_BUCKET_NAME = AwsBucketNameType()


class CanvasTokenType(click.ParamType):
    name = 'canvas token'

    def convert(self, value, param, ctx):
        value = value.strip()

        if not re.match(r'^[A-Za-z0-9]{64}$', value):
            self.fail("`{}' is not a valid Canvas token. Should be 64 "
                      "alphanumeric characters.".format(value), param, ctx)

        return value


CANVAS_TOKEN = CanvasTokenType()


class EmailParamType(click.ParamType):
    name = 'email'
    regex = r'^[_a-z0-9-]+(\.[_a-z0-9-]+)*@[a-z0-9-]+(\.[a-z0-9-]+)*' \
            r'(\.[a-z]{2,4})$'

    def convert(self, value, param, ctx):
        if re.match(EmailParamType.regex, value) is None:
            self.fail('%s is not a valid email address' % value, param, ctx)

        return value



import os
import re
import shlex
import tempfile
from fractions import Fraction

from ..submission import BrokenSubmissionError
from ..utils import run_process, PIPE, STDOUT, TimeoutExpired
from ..grades import PartGrade
from . import ThreadedGrader, Part

"""
Grade a native C assignment using Check, a C unit testing framework.
"""


class LibcheckTest(Part):
    __slots__ = ('name')

    REGEX_SUMMARY = re.compile(r'\d+%:\s+Checks:\s+(?P<total>\d+),\s+'
                               r'Failures:\s+(?P<failures>\d+),\s+'
                               r'Errors:\s+(?P<errors>\d+)')

    def __init__(self, name, disable_valgrind=False, valgrind_deduction=None):
        self.name = name
        self.disable_valgrind = disable_valgrind
        self.valgrind_deduction = Fraction(valgrind_deduction) \
            if valgrind_deduction is not None else None

    def description(self):
        return self.name

    @staticmethod
    def format_cmd(cmd, **kwargs):
        return [arg.format(**kwargs) for arg in cmd]

    @staticmethod
    def test_error_grade(message):
        return PartGrade(Fraction(0), deductions=('error',), log=message)

    def grade(self, path, grader):
        """Grade a single libcheck test"""

        grade = PartGrade(Fraction(1), deductions=[], log='')

        logfile_fp, logfile_path = tempfile.mkstemp(prefix='log-', dir=path)
        # Don't leak fds
        os.close(logfile_fp)
        logfile_basename = os.path.basename(logfile_path)

        run_cmd = self.format_cmd(grader.run_cmd, testcase=self.name,
                                  logfile=logfile_basename)
        process = run_process(run_cmd,
                              env={'CK_DEFAULT_TIMEOUT':
                                   str(grader.test_timeout)},
                              cwd=path, stdout=PIPE, stderr=STDOUT)

        if process.returncode != 0:
            return self.test_error_grade('tester exited with {} != 0:\n{}'
                                         .format(process.returncode,
                                                 process.stdout.decode()
                                                 if process.stdout is not None
                                                 else '(no output)'))
        try:
            # NOTE: drop any non-UTF8 characters (e.g. if a string is not
            # null-terminated)
            with open(logfile_path, 'r', errors='ignore') as logfile:
                logfile_contents = logfile.read()
                grade.log += logfile_contents

                summary = logfile_contents.splitlines()[-1]
                match = self.REGEX_SUMMARY.match(summary)
        except IOError as err:
            return self.test_error_grade('could not open log file: {}'
                                         .format(os.strerror(err.errno)))

        total = int(match.group('total'))
        failed = int(match.group('errors')) + int(match.group('failures'))
        grade.score *= Fraction(total - failed, total)

        if grader.valgrind_cmd is None or self.disable_valgrind:
            return grade

        grade.log += '\nValgrind\n--------\n'

        if grade.score < 1:
            valgrind_deduct = False
            grade.log += 'failed this test, so skipping valgrind...\n'
        else:
            # When running valgrind, we need to set CK_FORK=no, else
            # we're gonna get a ton of bogus reported memory leaks.
            # However, CK_FORK=no will break libcheck timeouts, so make
            # sure to handle timeouts here with subprocess
            try:
                valgrind_cmd = self.format_cmd(grader.valgrind_cmd,
                                               testcase=self.name,
                                               logfile=logfile_basename)
                valgrind_process = run_process(valgrind_cmd,
                                               stdout=PIPE,
                                               stderr=STDOUT,
                                               env={'CK_FORK': 'no'},
                                               cwd=path,
                                               timeout=grader.valgrind_timeout)
            except TimeoutExpired:
                grade.log += ('valgrind timed out after {} seconds, deducting '
                              'full valgrind penalty...\n').format(
                                  grader.valgrind_timeout)
                valgrind_deduct = True
            else:
                if valgrind_process.stdout:
                    grade.log += valgrind_process.stdout.decode()
                else:
                    grade.log += '(no output)'
                valgrind_deduct = valgrind_process.returncode != 0

        if valgrind_deduct:
            deduction = self.valgrind_deduction \
                if self.valgrind_deduction is not None \
                else grader.valgrind_deduction
            grade.deductions.append('valgrind')
            grade.score *= 1 - deduction

        return grade


class LibcheckGrader(ThreadedGrader):
    DEFAULT_BUILD_TIMEOUT = 15
    DEFAULT_TEST_TIMEOUT = 5
    DEFAULT_VALGRIND_TIMEOUT = 30

    def __init__(self, build_cmd, run_cmd, valgrind_cmd=None,
                 valgrind_deduction=None, build_timeout=None,
                 test_timeout=None, valgrind_timeout=None,
                 num_threads=None):
        super(LibcheckGrader, self).__init__(num_threads)

        self.build_cmd = shlex.split(build_cmd)
        self.run_cmd = shlex.split(run_cmd)

        self.valgrind_cmd = None \
            if valgrind_cmd is None else shlex.split(valgrind_cmd)
        self.valgrind_deduction = Fraction(1) \
            if valgrind_deduction is None else Fraction(valgrind_deduction)
        self.build_timeout = self.DEFAULT_BUILD_TIMEOUT \
            if build_timeout is None else build_timeout
        self.test_timeout = self.DEFAULT_TEST_TIMEOUT \
            if test_timeout is None else test_timeout
        self.valgrind_timeout = self.DEFAULT_VALGRIND_TIMEOUT \
            if valgrind_timeout is None else valgrind_timeout

    def list_prerequisites(self):
        return ['build-essential', 'check', 'valgrind', 'pkg-config']

    def grade_part(self, part, path, submission):
        return part.grade(path, self)

    def part_from_config_dict(self, config_dict):
        return LibcheckTest.from_config_dict(config_dict)

    def grade(self, submission, path, parts):
        try:
            process = run_process(self.build_cmd, cwd=path,
                                  timeout=self.build_timeout,
                                  stdout=PIPE,
                                  stderr=STDOUT)
        except TimeoutExpired:
            raise BrokenSubmissionError('timeout of {} seconds expired for '
                                        'build command'.format(self.timeout))

        if process.returncode != 0:
            raise BrokenSubmissionError(
                'build command exited with nonzero exit code {}'
                .format(process.returncode),
                verbose=process.stdout.decode() if process.stdout else None)

        return super(LibcheckGrader, self).grade(submission, path, parts)



# -*- coding: utf-8 -*-

"""Command-line interface to zucchini."""
import os
import re
import sys
import csv
import shutil
from functools import update_wrapper

import click

from .utils import EmailParamType, mkdir_p, CANVAS_URL, CANVAS_TOKEN, \
    AWS_ACCESS_KEY_ID, AWS_SECRET_ACCESS_KEY, AWS_BUCKET_NAME, \
    queue, run_thread
from .grading_manager import Grade, GradingManager
from .filter import FilterBuilder
from .zucchini import ZucchiniState
from .canvas import CanvasAPIError, CanvasNotFoundError, CanvasInternalError
from .constants import APP_NAME, USER_CONFIG, DEFAULT_SUBMISSION_DIRECTORY, \
                       SUBMISSION_FILES_DIRECTORY
from .submission import Submission
from .flatten import flatten, ArchiveError
from .gradescope import GradescopeMetadata, GradescopeAutograderOutput, \
                        GradescopeAutograderZip
from .loaders import CanvasArchiveLoader, GradescopeLoader
from .local_grading import LocalAutograderOutput

pass_state = click.make_pass_decorator(ZucchiniState)


def get_config_location():
    config_dir = click.get_app_dir(APP_NAME, force_posix=True, roaming=True)
    mkdir_p(config_dir)
    config_path = os.path.join(config_dir, USER_CONFIG)
    return config_dir, config_path


def load_config(zucchini_state):
    config_dir, config_path = get_config_location()

    try:
        with click.open_file(config_path, 'r') as config_file:
            zucchini_state.load_config_from_file(config_file, config_dir)
    except:  # noqa
        # TODO: Maybe better handling here, is it corrupt or nonexistent?
        click.echo("Need to set up configuration before doing any other work.")
        prompt_for_config(zucchini_state)


def prompt_for_config(zucchini_state):
    if not sys.stdin.isatty():
        raise click.ClickException('Cannot setup configuration interactively '
                                   'in a noninteractive session')

    click.echo("Zucchini will now set up your user configuration, overwriting "
               "any existing settings.")

    new_conf = {}
    new_conf['user-name'] = click.prompt('Your Name')
    new_conf['user-email'] = click.prompt('Your Email', type=EmailParamType())

    new_conf['canvas-url'] = click.prompt('Canvas URL (press enter to skip '
                                          'Canvas configuration)',
                                          type=CANVAS_URL, default='')
    if new_conf['canvas-url']:
        new_conf['canvas-token'] = click.prompt(
            'Canvas API token (To generate one, go to {}/profile/settings and '
            'choose "New Access Token")'
            .format(new_conf['canvas-url']), type=CANVAS_TOKEN)
    else:
        click.echo('Skipping canvas configuration...')
        new_conf['canvas-token'] = ''

    # for amazon
    new_conf['aws-access-key-id'] = click.prompt(
        'AWS Access Key ID (press enter to skip AWS configuration)',
        type=AWS_ACCESS_KEY_ID, default='')
    if new_conf['aws-access-key-id']:
        new_conf['aws-secret-access-key'] = click.prompt(
            'Amazon Secret Access Key', type=AWS_SECRET_ACCESS_KEY)
        new_conf['aws-s3-bucket-name'] = click.prompt(
            'Amazon S3 Bucket Name', type=AWS_BUCKET_NAME)
    else:
        click.echo('Skipping AWS configuration...')
        new_conf['aws-secret-access-key'] = ''
        new_conf['aws-s3-bucket-name'] = ''

    config_dir, config_path = get_config_location()
    zucchini_state.load_config_from_dict(new_conf, config_dir)

    with click.open_file(config_path, 'w') as config_file:
        zucchini_state.save_config_to_file(config_file)


def filter_options(canvas):
    """
    Decorator which adds common filtering options to a subcommand. If
    canvas=True, use only options supported by the Canvas API.
    """
    filter_options = {
        'student': click.option('-s', '--student', metavar='NAME',
                                multiple=True, help='Filter by student name'),
        'not_student': click.option('-S', '--not-student', metavar='NAME',
                                    multiple=True,
                                    help='Exclude student by name'),
        'broken': click.option('-b/-B', '--broken/--not-broken', default=None,
                               help='Filter for broken submissions'),
    }
    filter_add_methods = {'student': FilterBuilder.add_student_name,
                          'not_student': FilterBuilder.add_not_student_name,
                          'broken': FilterBuilder.add_broken}
    if canvas:
        filters = ('student', 'not_student')
    else:
        filters = ('student', 'not_student', 'broken')

    def decorator(func):
        def replacement(*args, **kwargs):
            filter_builder = FilterBuilder.new_canvas() \
                             if canvas else FilterBuilder.new_meta()
            for filter_ in filters:
                terms = kwargs.pop(filter_)
                add_method = filter_add_methods[filter_]
                if terms is not None:
                    if isinstance(terms, (list, tuple)):
                        for term in terms:
                            add_method(filter_builder, term)
                    else:
                        add_method(filter_builder, terms)

            kwargs['filter'] = filter_builder
            func(*args, **kwargs)

        for filter_ in reversed(filters):
            replacement = filter_options[filter_](replacement)

        # Without update_wrapper() here, click gets confused and puts
        # a replacement subcommand in the help output
        return update_wrapper(replacement, func)

    return decorator


def need_config(func):
    """
    Marks that this command requires configuration and should prompt for
    it if necessary.
    """

    def replacement(*args, **kwargs):
        zucchini_state = args[0]
        load_config(zucchini_state)
        func(*args, **kwargs)

    # Without update_wrapper() here, click gets confused and puts
    # a replacement subcommand in the help output
    return update_wrapper(replacement, func)


@click.group()
@click.option('-a', '--assignment', default='.',
              help="Path of the directory containing the Zucchini assignment.",
              type=click.Path(exists=True, file_okay=False, dir_okay=True,
                              writable=True, readable=True,
                              resolve_path=True))
@click.pass_context
def cli(ctx, assignment):
    """
    zucchini, a fun autograder for the whole family.

    \b
    example workflow:
        $ zucc farm add https://some/git/repo/url.git my-farm
        $ zucc init my-farm/my-assignment
        $ cd my-assignment
        $ zucc load canvas
        $ zucc grade
        $ zucc export canvas

    documentation: https://zucchini.readthedocs.io/
    """
    ctx.obj = ZucchiniState(assignment)


@cli.command()
@pass_state
def setup(state):
    """Runs setup again"""
    prompt_for_config(state)


@cli.command()
@pass_state
@need_config
def update(state):
    """Update all farms."""
    # TODO: Add support for single-farm listing
    click.echo("Updating farms...")
    state.farm_manager.update_all_farms()
    click.echo("Successfully updated all farms.")
    # TODO: Add stats about # configurations etc


@cli.command('list')
@pass_state
@need_config
def list_assignments(state):
    """Update all farms and list downloadable assignments."""
    # TODO: Add support for single-farm listing
    # TODO: also potentially do this by invoking the update cmd
    click.echo("Updating farms...")
    state.farm_manager.update_all_farms()
    click.echo("Successfully updated all farms.\n")

    click.echo("Available assignments:")
    assignments = state.farm_manager.list_farm_assignments()
    click.echo("\n".join(["%s: %s" % x for x in assignments]))


@cli.command()
@click.argument('assignment-name')
@click.option('-t', '--target', default='.',
              help="Path of the directory to clone the zucchini assignment"
                   "folder into.",
              type=click.Path(exists=True, file_okay=False, dir_okay=True,
                              writable=True, readable=True,
                              resolve_path=True))
@pass_state
@need_config
def init(state, assignment_name, target):
    """Configure an assignment for grading."""
    state.farm_manager.clone_farm_assignment(assignment_name, target)
    click.echo("Successfully initialized %s into %s." % (assignment_name,
                                                         target))


@cli.group()
@click.option('-t', '--to-dir', default=DEFAULT_SUBMISSION_DIRECTORY,
              help="Path of the directory in which to put submissions loaded.",
              type=click.Path(file_okay=False, dir_okay=True,
                              writable=True, readable=True,
                              resolve_path=True))
@pass_state
def load(state, to_dir):
    """Load student submissions."""

    state.submission_dir = to_dir


@load.command('path')
@click.argument('path',
                type=click.Path(file_okay=True, dir_okay=True, readable=True,
                                resolve_path=True))
@click.option('-n', '--student-name', metavar='NAME', default=None,
              help='Name of student. Will prompt if missing')
@click.option('-d', '--default-student-name', is_flag=True,
              help="Use the name of the imported object as student name.")
@click.option('--max-archive-size', type=int, metavar='BYTES',
              help='maximum size of archive to extract')
@pass_state
@need_config
def load_path(state, path, student_name, default_student_name,
              max_archive_size):
    """
    Load a submission from a filesystem path.

    If path is a directory, will use all files in the directory as the
    submission files. If path is a file, will use only that file.
    Flattens submission files (extracts archives) after copying.
    """

    if student_name is None:
        # Try to guess the student's name based on the directory/file name
        name_pieces = re.split(r'[-_\s,]+', os.path.basename(path))
        name_guess = ' '.join(name.capitalize() for name in name_pieces)

        if default_student_name:
            student_name = name_guess
        else:
            student_name = click.prompt('Student name', default=name_guess)

    base_dir = os.path.join(state.submission_dir, student_name)
    # Remove submission if it already exists
    shutil.rmtree(base_dir, ignore_errors=True)

    mkdir_p(base_dir)
    # Wait to create files_dir because copytree() below expects the
    # destination directory not to exist
    files_dir = os.path.join(base_dir, SUBMISSION_FILES_DIRECTORY)

    if os.path.isdir(path):
        shutil.copytree(path, files_dir)
    else:
        mkdir_p(files_dir)
        destfile = os.path.join(files_dir, os.path.basename(path))
        shutil.copy(path, destfile)

    flatten(files_dir, max_archive_size=max_archive_size)

    # Create initial meta.json in submission dir
    submission = Submission.load_from_empty_dir(
        state.get_assignment(), base_dir, student_name=student_name,
        graded=False)
    submission.initialize_metadata()


@load.command('gradescope')
@click.argument('export-zipfile',
                type=click.Path(file_okay=True, dir_okay=True, readable=True,
                                resolve_path=True))
@pass_state
def load_gradescope(state, export_zipfile):
    """Load submissions from Gradescope"""
    with GradescopeLoader(export_zipfile) as loader:
        with click.progressbar(list(loader.submissions.items())) as bar:
            for submission_id, student_name in bar:
                base_dir = os.path.join(state.submission_dir, student_name)
                # Remove submission if it already exists
                shutil.rmtree(base_dir, ignore_errors=True)

                files_dir = os.path.join(base_dir, SUBMISSION_FILES_DIRECTORY)
                mkdir_p(files_dir)

                loader.extract_files(submission_id, files_dir)

                # Create initial meta.json in submission dir
                submission = Submission.load_from_empty_dir(
                    state.get_assignment(), base_dir,
                    student_name=student_name, graded=False, id=submission_id)
                submission.initialize_metadata()


def choose_section(sections, section=None):
    """
    Prompt and return a CanvasSection object or None
    """

    section_chosen = None

    # Find a section matching criteria (either id or substring of section name)
    while True:
        error_message = None

        # If this is not their first attempt, print an error describing
        # why we're prompting again
        if section is not None:
            if section == 'all':
                break

            # First, try to find a match by id
            try:
                section_id = int(section)
                id_matches = [s for s in sections if section_id == s.id]

                if id_matches:
                    # Assume a section id is unique
                    section_chosen, = id_matches
                    break
            except ValueError:
                # Not an integer
                pass

            # Now, try to find a match by name
            name_matches = [s for s in sections if section in s.name.lower()]

            if len(name_matches) == 1:
                section_chosen, = name_matches
                break
            elif len(name_matches) > 1:
                error_message = 'More than one section matches. Try again? ' \
                                '(Canvas is an extremely good website and ' \
                                'allows duplicate section names, so you may ' \
                                'have to supply an id.)'
            else:
                error_message = 'No sections match. Try again?'

        click.echo('List of sections:')
        for s in sections:
            click.echo(str(s))

        # Print the error message _after_ the list of sections. Even if
        # there are tons of sections, we still want the user to see the
        # error message.
        if error_message:
            click.echo(error_message)

        section = click.prompt('Choose a section (name or id)',
                               default='all', type=lambda s: s.lower())

    return section_chosen


def canvas_setup(state, section):
    """
    Check configuration for the Canvas API.

    Return a tuple of (CanvasAPI instance, course id, assignment id,
    CanvasSection instance or None).
    """

    course_id = state.get_assignment().canvas_course_id
    if course_id is None:
        raise click.ClickException('Need to configure canvas in assignment '
                                   'config')
    api = state.canvas_api()
    try:
        sections = tuple(api.list_sections(course_id))
    except CanvasNotFoundError:
        raise click.ClickException('Canvas reports no course with id {}'
                                   .format(course_id))
    except CanvasInternalError:
        raise click.ClickException('Canvas reported an internal error (5xx '
                                   'status code). Try again later?')
    except CanvasAPIError as err:
        raise click.ClickException(str(err))

    if not sections:
        raise click.ClickException('No sections, so no students! Bailing out')

    section_chosen = choose_section(sections, section)
    assignment_id = state.get_assignment().canvas_assignment_id

    return (api, course_id, assignment_id, section_chosen)


@load.command('canvas')
@click.option('--section', '-e', type=lambda s: s.lower(), metavar='SECTION',
              help='section id, substring of section name, or "all"')
@click.option('--max-archive-size', type=int, metavar='BYTES',
              help='maximum size of archive to extract')
@filter_options(canvas=True)
@pass_state
@need_config
def load_canvas(state, section, max_archive_size, filter):
    """Load submissions from Canvas"""

    api, course_id, assignment_id, section_chosen = \
        canvas_setup(state, section)

    # If they specified "all", use all submissions in the course,
    # otherwise use just those from one section
    if section_chosen is None:
        submissions = api.list_submissions(course_id, assignment_id)
    else:
        submissions = api.list_section_submissions(section_chosen.id,
                                                   assignment_id)

    click.echo('Downloading submissions from Canvas...')
    # Need to iterate over the list of submissions so that click can
    # call len(iterator) to know how to progress the progress bar
    with click.progressbar(list(submissions)) as bar:
        for canvas_submission in bar:
            if not filter(canvas_submission):
                continue

            student_name = canvas_submission.user.sortable_name
            base_dir = os.path.join(state.submission_dir, student_name)
            # Remove submission if it already exists
            shutil.rmtree(base_dir, ignore_errors=True)

            files_dir = os.path.join(base_dir, SUBMISSION_FILES_DIRECTORY)
            mkdir_p(files_dir)

            if canvas_submission.no_submission():
                error = 'No submission!'
            else:
                error = None
                canvas_submission.download(files_dir)
                try:
                    flatten(files_dir, max_archive_size=max_archive_size)
                except ArchiveError as err:
                    error = str(err)

            # Create initial meta.json in submission dir
            submission = Submission.load_from_empty_dir(
                state.get_assignment(), base_dir, graded=False,
                student_name=student_name, id=canvas_submission.user_id,
                seconds_late=canvas_submission.seconds_late, error=error)
            submission.initialize_metadata()


@load.command('canvas-archive')
@click.argument('bulk-zipfile',
                type=click.Path(file_okay=True, dir_okay=True, readable=True,
                                resolve_path=True))
@click.option('--section', '-e', type=lambda s: s.lower(), metavar='SECTION',
              help='section id, substring of section name, or "all"')
@click.option('--max-archive-size', type=int, metavar='BYTES',
              help='maximum size of archive to extract')
@filter_options(canvas=True)
@pass_state
@need_config
def load_canvas_archive(state, bulk_zipfile, section, max_archive_size,
                        filter):
    """
    Load submissions from Canvas bulk download
    """

    api, course_id, assignment_id, section_chosen = \
        canvas_setup(state, section)

    # TODO: Implement extraction for all students
    if section_chosen is None:
        raise click.ClickException('Please specify a section to extract. '
                                   'Extracting all is not implemented yet')
    else:
        students = list(
            api.list_section_students(course_id, section_chosen.id))

    click.echo('Extracting bulk Canvas submissions...')

    with CanvasArchiveLoader(bulk_zipfile) as loader:
        with click.progressbar(students) as bar:
            for student in bar:
                base_dir = os.path.join(state.submission_dir,
                                        student.sortable_name)
                # Remove submission if it already exists
                shutil.rmtree(base_dir, ignore_errors=True)

                files_dir = os.path.join(base_dir, SUBMISSION_FILES_DIRECTORY)
                mkdir_p(files_dir)

                if not loader.has_submission(student.id):
                    error = 'No submission!'
                else:
                    error = None
                    loader.extract_files(student.id, files_dir)
                    try:
                        flatten(files_dir, max_archive_size=max_archive_size)
                    except ArchiveError as err:
                        error = str(err)

                # Create initial meta.json in submission dir
                submission = Submission.load_from_empty_dir(
                    state.get_assignment(), base_dir,
                    student_name=student.sortable_name, graded=False,
                    id=student.id, error=error)
                submission.initialize_metadata()


@cli.command('grade-submission')
@click.argument('submission-path',
                type=click.Path(exists=True, file_okay=False, dir_okay=True,
                                readable=True))
@pass_state
def grade_submission(state, submission_path):
    """
    Grade a single submission not `zucc load`ed.

    Do NOT use this to (re-)grade some submissions in submissions/
    loaded by `zucc load ...`. Instead, use `zucc grade -s "Lin,
    Michael"'. However, if you need to grade a single submission from
    somewhere else — just submission files, no meta.json or anything —
    use this, passing it the path to the submission files.
    """

    submission = Submission.load_from_raw_files(state.get_assignment(),
                                                submission_path)
    grade = Grade(state.get_assignment(), submission)
    grade.grade()
    grade.dump_component_grades(sys.stdout)


def print_grades(grades, grader_name):
    """Display grades, an iterable of Grade instances, in a pager."""
    grades = sorted(grades,
                    key=lambda grade: grade.student_name())
    # Length of longest name
    max_name_len = max(len(grade.student_name()) for grade in grades)

    grade_report = '\n'.join(
        '{:<{max_name_len}}\t{}\t{}'.format(
            grade.student_name(),
            # If the assignment is ungradable, show a 0 instead of
            # "(ungraded)"
            grade.score() if not grade.gradable() or grade.graded()
            else '(ungraded)',
            grade.breakdown(grader_name)
            if not grade.gradable() or grade.graded() else '',
            max_name_len=max_name_len)
        for grade in grades)
    click.echo_via_pager('grade report:\n\n' + grade_report)


@cli.command()
@click.option('-f', '--from-dir', default=DEFAULT_SUBMISSION_DIRECTORY,
              help="Path of the directory to read submissions from.",
              type=click.Path(exists=True, file_okay=False, dir_okay=True,
                              writable=True, readable=True,
                              resolve_path=True))
@filter_options(canvas=False)
@pass_state
@need_config
def grade(state, from_dir, filter):
    """Grade submissions."""
    # At this point, the assignment object is loaded. We need a GradingManager
    grading_manager = GradingManager(state.get_assignment(), from_dir, filter)

    if not grading_manager.submission_count():
        raise click.ClickException('no submissions match the filter given!')

    click.echo('Grading submissions...')

    if grading_manager.has_interactive():
        if grading_manager.has_noninteractive():
            # Since we have both interactive an non-interactive
            # components, we want to let the user grade the interactive
            # components while the non-interactive components are
            # running.
            grade_queue = queue.Queue()
            thread = run_thread(grading_manager.grade, (False,), grade_queue)

            click.echo('First, grading interactive components...')
            grades = list(grading_manager.grade(interactive=True))
            click.echo('Finishing off grading noninteractive components...')

            with click.progressbar(length=grading_manager.submission_count()) \
                    as bar:
                for i in bar:
                    noninter_grade = grade_queue.get()
                    if isinstance(noninter_grade, Exception):
                        raise noninter_grade
                    grades[i].update(noninter_grade)

            # Should have exited by now, but just in case
            thread.join()
        else:
            # If all the components are interactive, just grade them
            # here in the main thread
            grades = list(grading_manager.grade())
    else:
        # If all components are noninteractive, just do the progress bar
        # in the main thread here.
        with click.progressbar(grading_manager.grade(),
                               grading_manager.submission_count()) as bar:
            grades = list(bar)

    # Need to do this now to handle non-interactive and interactive
    # running in parallel earlier
    for grade in grades:
        grade.write_grade()
        grade.generate_gradelog()
    print_grades(grades, state.user_name)


@cli.command('show-grades')
@click.option('-f', '--from-dir', default=DEFAULT_SUBMISSION_DIRECTORY,
              help="Path of the directory to read submissions from.",
              type=click.Path(exists=True, file_okay=False, dir_okay=True,
                              writable=True, readable=True,
                              resolve_path=True))
@filter_options(canvas=False)
@pass_state
@need_config
def show_grades(state, from_dir, filter):
    """Print the grade for all submissions."""
    grading_manager = GradingManager(state.get_assignment(), from_dir, filter)

    if not grading_manager.submission_count():
        raise click.ClickException('no submissions match the filter given!')

    print_grades(grading_manager.grades(), state.user_name)


@cli.group()
@click.option('-f', '--from-dir', default=DEFAULT_SUBMISSION_DIRECTORY,
              help="Path of the directory to read submissions from.",
              type=click.Path(exists=True, file_okay=False, dir_okay=True,
                              writable=True, readable=True,
                              resolve_path=True))
@filter_options(canvas=False)
@pass_state
def export(state, from_dir, filter):
    """Export grades for uploading."""

    grading_manager = GradingManager(state.get_assignment(), from_dir, filter)

    if not grading_manager.submission_count():
        raise click.ClickException('no submissions match the filter given!')

    state.grades = [grade for grade in grading_manager.grades()
                    if grade.grade_ready()]


@export.command('csv')
@click.option('--out-file', '-o', help='Where to write the CSV. Default is '
                                       'stdout',
              type=click.Path(file_okay=True, dir_okay=False,
                              resolve_path=True))
@pass_state
@need_config
def export_csv(state, out_file=None):
    """Export grades to an Excel CSV."""

    if out_file is None:
        csvfile = sys.stdout
    else:
        csvfile = open(out_file, 'w')

    try:
        writer = csv.writer(csvfile)
        for grade in state.grades:
            writer.writerow([grade.student_name(), grade.score(),
                             grade.breakdown(state.user_name)])
    finally:
        if out_file is not None:
            csvfile.close()


@export.command('canvas-grades')
@pass_state
@need_config
def export_canvas_grades(state):
    """Upload grades to Canvas."""

    api = state.canvas_api()
    course_id = state.get_assignment().canvas_course_id
    assignment_id = state.get_assignment().canvas_assignment_id

    if None in (course_id, assignment_id):
        raise click.ClickException('Need to configure canvas in assignment '
                                   'config')

    click.echo('Uploading grades to canvas...')
    with click.progressbar(state.grades) as bar:
        for grade in bar:
            # Submissions not from canvas won't have an id set, so skip them
            if grade.student_id() is not None:
                api.set_submission_grade(course_id, assignment_id,
                                         grade.student_id(), grade.score())


@export.command('canvas-comments')
@click.option('--gradelog-upload', '-g', help='Where to upload Gradelog file',
              type=click.Choice(['none', 'canvas', 's3']))
@pass_state
@need_config
def export_canvas_comments(state, gradelog_upload):
    """Add Canvas submission comments with gradelog and breakdown"""
    if gradelog_upload is None:
        click.echo('gradelog_upload flag is mandatory')
        return

    click.echo('Gradelog upload type: ' + gradelog_upload)

    api = state.canvas_api()
    course_id = state.get_assignment().canvas_course_id
    assignment_id = state.get_assignment().canvas_assignment_id

    if None in (course_id, assignment_id):
        raise click.ClickException('Need to configure canvas in assignment '
                                   'config')

    click.echo('Uploading submission comments to canvas...')
    with click.progressbar(state.grades) as bar:
        for grade in bar:
            # Submissions not from canvas won't have an id set, so skip them
            if grade.student_id() is not None:
                breakdown = grade.breakdown(state.user_name)
                gradelog_hash = grade.get_gradelog_hash()
                gradelog_path = grade.get_gradelog_path()
                student_id = grade.student_id()

                if gradelog_upload == 'none':
                    api.add_submission_comment(
                        course_id, assignment_id, student_id,
                        breakdown,
                        None
                    )
                elif gradelog_upload == 'canvas':
                    api.add_submission_comment(
                        course_id, assignment_id, student_id,
                        breakdown + '\n\n' + gradelog_hash,
                        [(gradelog_path, 'text/plain')]
                    )
                elif gradelog_upload == 's3':
                    file_url = state.get_amazon_api().upload_file_s3(
                        gradelog_path, 'text/plain'
                    )
                    api.add_submission_comment(
                        course_id, assignment_id, student_id,
                        breakdown + "\n\n" + gradelog_hash + '\n\n' + file_url,
                        None
                    )


@cli.group()
def farm():
    """Manage zucchini farms."""
    pass


@farm.command('add')
@click.argument('farm-url')
@click.argument('farm-name')
@pass_state
@need_config
def add_farm(state, farm_url, farm_name):
    state.farm_manager.add_farm(farm_url, farm_name)
    click.echo("Successfully added farm %s" % farm_name)


@farm.command('recache')
@click.argument('farm-name')
@pass_state
@need_config
def recache_farm(state, farm_name):
    state.farm_manager.recache_farm(farm_name)
    click.echo("Successfully recached farm %s" % farm_name)


@farm.command('list')
@pass_state
@need_config
def list_farms(state):
    farms = state.farm_manager.list_farms()
    click.echo("\n".join(farms))


@farm.command('remove')
@click.argument('farm-name')
@pass_state
@need_config
def remove_farm(state, farm_name):
    state.farm_manager.remove_farm(farm_name)
    click.echo("Successfully removed farm %s" % farm_name)


@cli.group('canvas-api')
def canvas_api():
    """Talk to the Canvas API (low-level interface)."""
    pass


@canvas_api.command('courses')
@pass_state
@need_config
def canvas_api_courses(state):
    """List Canvas courses"""

    api = state.canvas_api()
    for course in api.list_courses():
        click.echo(str(course))


@canvas_api.command('assignments')
@click.argument('course-id')
@pass_state
@need_config
def canvas_api_assignments(state, course_id):
    """List assignments in a Canvas course"""

    api = state.canvas_api()
    for assign in api.list_assignments(course_id):
        click.echo(str(assign))


@canvas_api.command('sections')
@click.argument('course-id')
@pass_state
@need_config
def canvas_api_sections(state, course_id):
    """List sections in a Canvas course"""

    api = state.canvas_api()
    for section in api.list_sections(course_id):
        click.echo(str(section))


@canvas_api.command('section-students')
@click.argument('course-id')
@click.argument('section-id')
@pass_state
@need_config
def canvas_api_section_students(state, course_id, section_id):
    """List students in a Canvas section"""

    api = state.canvas_api()
    for student in api.list_section_students(course_id, section_id):
        click.echo(str(student))


@canvas_api.command('submissions')
@click.argument('course-id')
@click.argument('assignment-id')
@pass_state
@need_config
def canvas_api_submissions(state, course_id, assignment_id):
    """List submissions for a Canvas assignment"""

    api = state.canvas_api()
    for submission in api.list_submissions(course_id, assignment_id):
        click.echo(str(submission))


@canvas_api.command('download')
@click.argument('course-id')
@click.argument('assignment-id')
@click.argument('user-id')
@click.argument('dest-directory')
@pass_state
@need_config
def canvas_api_download(state, course_id, assignment_id, user_id,
                        dest_directory):
    """Download a submission"""

    api = state.canvas_api()
    submission = api.get_submission(course_id, assignment_id, user_id)
    submission.download(dest_directory)


@canvas_api.command('grade')
@click.argument('course-id')
@click.argument('assignment-id')
@click.argument('user-id')
@click.argument('grade')
@click.option('--comment', help='Add TEXT as a new grading comment')
@pass_state
@need_config
def canvas_api_grade(state, course_id, assignment_id, user_id, grade,
                     comment=None):
    """
    Grade a submission.

    Warning: supplying --comment adds a new grade comment; it does not update
    existing comments. Thus, every time you run `zucc canvas grade ...
    --comment hello', you will add a brand new comment "hello." This is because
    the canvas API cannot edit or delete submission comments; it can only add
    them.
    """

    api = state.canvas_api()
    api.set_submission_grade(course_id, assignment_id, user_id, grade, comment)


@cli.group('gradescope')
def gradescope_():
    """Handy scripts for Gradescope autograding."""
    pass


@gradescope_.command('bridge')
@click.argument('metadata-path')
@pass_state
def gradescope_bridge(state, metadata_path):
    """
    Convert component grades to gradescope JSON.

    Bridge the gap between component results provided obtained with
    `zucc grade-submission' (plus Gradescope submission metadata) and a
    Gradescope JSON result. Will read component grades JSON from stdin
    and spit out Gradescope JSON on stdout.

    \b
    Usual usage:
    zucc grade-submission /autograder/submission \\
        | zucc gradescope bridge /autograder/submission_metadata.json \\
        >/autograder/results/results.json
    """

    metadata = GradescopeMetadata.from_json_path(metadata_path)
    assignment = state.get_assignment()

    late_deltatime = metadata.submission_date - metadata.due_date
    seconds_late = max(int(late_deltatime.total_seconds()), 0)

    submission = Submission.load_from_component_grades_json(
        assignment, seconds_late=seconds_late, component_grades_fp=sys.stdin)
    submission.student_name = metadata.student_name
    grade = Grade(assignment, submission, max_score=metadata.total_points)

    output = GradescopeAutograderOutput.from_grade(grade)
    output.to_json_stream(sys.stdout)


@gradescope_.command('zip')
@click.option('--out-path', '-o', metavar='ZIP_FILE', default='autograder.zip',
              help='path to zip file to generate')
@click.option('--wheel-path', metavar='WHEEL_PATH', default=None,
              help='path to zucchini wheel to install into the image. '
                   'useful for development. default behavior: install the '
                   'same version of zucchini as this version from pypi')
@pass_state
def gradescope_zip(state, out_path, wheel_path):
    """
    Generate a gradescope autograder .zip.

    Creates an autograder zip file at ZIP_FILE for this zucchini
    assignment ready to upload to Gradescope.
    """
    autograder_zip = GradescopeAutograderZip(
        path=state.assignment_directory,
        prerequisites=state.get_assignment().list_prerequisites(),
        extra_setup_commands=state.get_assignment()
                                  .list_extra_setup_commands(),
        needs_display=state.get_assignment().needs_display(),
        wheel_path=wheel_path)
    autograder_zip.write_zip(out_path)


@cli.command('flatten')
@click.argument('dir-path')
@click.option('--max-archive-size', type=int, metavar='BYTES',
              help='maximum size of archive to extract')
@pass_state
def flatten_(self, dir_path, max_archive_size):
    """
    Flatten archives in a directory.

    Not as good, but behaves similarly to Marie's classic
    SubmissionFix.py script. Give it a directory and it will extract all
    of the archives in the top level, removing a common directory prefix
    from them if it exists. Checks for malicious archives like zipbombs
    and forged archive filenames.
    """
    flatten(dir_path, max_archive_size=max_archive_size)


@cli.group('local')
def local_():
    """Handy scripts for local autograding."""
    pass


@local_.command('print')
@pass_state
def local_print(state):
    """
    Convert component grades to printed output.

    Take the component results obtained with `zucc grade-submission' and
    output it to the command line. Used to enable local autograding
    without having to setup the zucchini config.

    \b
    Usual usage:
    zucc grade-submission /autograder/submission \\
        | zucc local print
    """

    assignment = state.get_assignment()

    submission = Submission.load_from_component_grades_json(
        assignment, component_grades_fp=sys.stdin)
    grade = Grade(assignment, submission)
    output = LocalAutograderOutput.from_grade(grade)

    click.echo_via_pager(output)


if __name__ == "__main__":
    cli()



import os
import shutil
import re

import git
import yaml

from .utils import mkdir_p
from .constants import FARM_ASSIGNMENT_NAME_REGEX, FARM_IDENTIFIER_FILE


class FarmAssignment(object):
    """A farm assignment is an assignment on a farm that can be downloaded.
    The file here is a simple .yml file that contains just a name, maintainer,
    and a Git URL for the assignment."""

    def __init__(self, name, maintainer, url):
        self.name = name
        self.maintainer = maintainer
        self.url = url

    def clone(self, path):
        return git.Repo.clone_from(url=self.url, to_path=path)
        # TODO: Why return a repo? We need success info

    def __str__(self):
        return "%s (Maintainer: %s)" % (self.name, self.maintainer)


class Farm(object):
    def __init__(self, path):
        self.path = path

        self.repo = git.Repo(self.path)
        self.farm_assignments = dict()

        self._parse_assignments()

    def _parse_assignments(self):
        self.farm_assignments = dict()

        # Let's read the directory structure here
        for root, subdirs, files in os.walk(self.path):
            for file_name in files:
                match = re.match(FARM_ASSIGNMENT_NAME_REGEX, file_name)

                if match is not None:  # This might be an assignment
                    try:
                        file_path = os.path.join(root, file_name)
                        with open(file_path, 'r') as config_file:
                            config = yaml.safe_load(config_file)
                            assignment = FarmAssignment(**config)

                            slug = os.path.splitext(
                                os.path.relpath(file_path, self.path))[0]
                            self.farm_assignments[slug] = assignment
                    except:  # noqa
                        # TODO: Do something here: the load failed
                        pass

    def get_farm_assignment_by_name(self, assignment_name):
        return self.farm_assignments[assignment_name]

    def update(self):
        """Call Git Pull on the repository to update its contents"""
        self.repo.remotes.origin.pull()
        # TODO: Get some info here about the fetch status
        self._parse_assignments()

    def list_assignments(self):
        return sorted(self.farm_assignments.items(), key=lambda x: x[0])


class FarmManager(object):
    def __init__(self, farm_root):
        self.root = farm_root

        # Generate our farms dir if it doesnt exist yet
        mkdir_p(self.root)

        self.farms = {}
        self._parse_farms()

    def _parse_farms(self):
        self.farms = {}

        farms = [name for name in os.listdir(self.root) if
                 os.path.isdir(os.path.join(self.root, name))]
        self.farms = {name: Farm(os.path.join(self.root, name)) for name in
                      farms}

    def get_farm_by_name(self, name):
        return self.farms[name]

    def get_path_for_farm_name(self, farm_name):
        return os.path.join(self.root, farm_name)

    def farm_exists(self, farm_name):
        return os.path.exists(self.get_path_for_farm_name(farm_name))

    def list_farms(self):
        return sorted(self.farms.keys())

    def list_farm_assignments(self):
        assignments = []
        for farm_name, farm in self.farms.items():
            assignments += [("%s/%s" % (farm_name, x[0]), x[1]) for x in
                            farm.list_assignments()]

        return sorted(assignments, key=lambda x: x[0])

    def list_farm_assignments_by_farm(self, farm_name):
        farm = self.get_farm_by_name(farm_name)
        return farm.list_assignments()

    def clone_farm_assignment(self, assignment_name, path):
        parsed_name = assignment_name.split('/', 1)

        if len(parsed_name) != 2:
            raise ValueError("You need to specify assignment names in"
                             "farm/assignment format.")

        farm, assignment = parsed_name

        clone_path = os.path.join(path, assignment.split('/')[-1])
        self.get_farm_by_name(farm).get_farm_assignment_by_name(
            assignment).clone(clone_path)

    def add_farm(self, farm_url, farm_name):
        if self.farm_exists(farm_name):
            raise ValueError("A farm by this name already exists.")

        path = self.get_path_for_farm_name(farm_name)
        repo = git.Repo.clone_from(url=farm_url, to_path=path)

        if repo is None:
            raise ValueError("Could not add farm: is the URL valid?")

        if not os.path.exists(os.path.join(path, FARM_IDENTIFIER_FILE)):
            shutil.rmtree(path)
            raise ValueError("The given git repo is not a valid zucchini "
                             "farm.")

        self._parse_farms()

    def update_farm(self, farm_name):
        farm = self.get_farm_by_name(farm_name)
        farm.update()

    def update_all_farms(self):
        for farm in self.farms.values():
            farm.update()

    def recache_farm(self, farm_name):
        # TODO: FIX THIS
        # farm = self.get_farm_by_name(farm_name)

        # farm_url = farm.repo.remotes.origin.urls[0] <--- the problem

        # self.remove_farm(farm_name)
        # self.add_farm(farm_url, farm_name)

        pass

    def remove_farm(self, farm_name):
        farm = self.get_farm_by_name(farm_name)
        shutil.rmtree(farm.path)

        self._parse_farms()



APP_NAME = "zucchini"
USER_CONFIG = "user.yml"
FARM_DIRECTORY = "farms"
FARM_IDENTIFIER_FILE = ".zucchini"
FARM_ASSIGNMENT_NAME_REGEX = '^(.*).yml$'
ASSIGNMENT_CONFIG_FILE = "zucchini.yml"
ASSIGNMENT_FILES_DIRECTORY = "grading-files"
DEFAULT_SUBMISSION_DIRECTORY = "submissions"
SUBMISSION_META_FILE = "meta.json"
SUBMISSION_FILES_DIRECTORY = "files"
SUBMISSION_GRADELOG_FILE = "gradelog.txt"
READ_BUFFER_SIZE = 65536



import re
from fractions import Fraction

from ..utils import run_process, PIPE, STDOUT
from ..grades import PartGrade
from . import ThreadedGrader, Part


class LC3ToolsTest(Part):
    __slots__ = ('name')

    def __init__(self, name):
        self.name = name

    def description(self):
        return self.name

    @staticmethod
    def format_cmd(cmd, **kwargs):
        return [arg.format(**kwargs) for arg in cmd]

    @staticmethod
    def test_error_grade(message):
        return PartGrade(Fraction(0), deductions=('error',), log=message)

    def grade(self, path, grader):
        grade = PartGrade(Fraction(1), log='')

        run_cmd = self.format_cmd(grader.cmdline, testcase=self.name)

        process = run_process(run_cmd, cwd=path, stdout=PIPE, stderr=STDOUT)

        if process.returncode != 0:
            return self.test_error_grade('tester exited with {} != 0:\n{}'
                                         .format(process.returncode,
                                                 process.stdout.decode()
                                                 if process.stdout is not None
                                                 else '(no output)'))

        out_contents = process.stdout.decode()
        out_contents = re.sub(r'\(\+.*pts\)', '', out_contents)
        results = "".join(out_contents.strip().splitlines(keepends=True)[:-1])
        grade.log += results

        try:
            summary = out_contents.splitlines()[-1]
            score = summary.replace("/", " ").split()[3:5]
            score[0] = Fraction(float(score[0]))
            score[1] = Fraction(float(score[1]))
            grade.score *= Fraction(score[0], score[1])
        except (ValueError, IndexError):
            return self.test_error_grade('Could not assemble file: \n{}'
                                         .format(process.stdout.decode()
                                                 if process.stdout is not None
                                                 else '(no output)'))
        return grade


class LC3ToolsGrader(ThreadedGrader):
    """
    Run a LC3Tools grader and collect the results.
    """

    DEFAULT_TIMEOUT = 30

    def __init__(self, test_file, asm_file, timeout=None, num_threads=None):

        super(LC3ToolsGrader, self).__init__(num_threads)
        self.test_file = test_file
        self.asm_file = asm_file

        self.cmdline = ["./" + self.test_file, self.asm_file,
                        '--test-filter={testcase}', "--tester-verbose",
                        "--asm-print-level=3"]

        self.timeout = self.DEFAULT_TIMEOUT \
            if timeout is None else timeout

    def list_prerequisites(self):
        return []

    def part_from_config_dict(self, config_dict):
        return LC3ToolsTest.from_config_dict(config_dict)

    def grade_part(self, part, path, submission):
        return part.grade(path, self)

    def grade(self, submission, path, parts):

        return super(LC3ToolsGrader, self).grade(submission, path, parts)



"""
Utilities for local autograding.
"""


class LocalAutograderOutput:
    """
    Take a Grade object and convert that to an output string that will be
    printed into the terminal
    """

    @classmethod
    def from_grade(cls, grade):
        computed_grade = grade.computed_grade()
        output = []
        total_got = 0.0
        total_max = 0.0
        any_errors = False
        for component in computed_grade.components:
            total_got += component.points_got
            total_max += component.points_possible
            if component.error:
                any_errors = True
                output.append("ERROR: %-45s %-15s" %
                              (component.name, component.error))
                output.append("Details: %-45s" % component.error_verbose)
            else:
                for part in component.parts:
                    points_got = grade.to_float(part.points_got)
                    points_max = grade.to_float(part.points_possible)
                    points = '{:.2f}/{:.2f}'.format(points_got, points_max)
                    if part.points_got < part.points_possible:
                        output.append("TEST: %-45s %-15s (%s)" %
                                      (part.name, "FAILED", points))
                        output.append(part.log)
                    else:
                        output.append("TEST: %-45s %-15s (%s)" %
                                      (part.name, "PASSED", points))
        score_percentage = 100.0 * total_got / total_max
        score = 'Total score: {:.2f}%'.format(score_percentage)
        output.append(score)
        if any_errors:
            output.append('Some errors occurred; the score above may not be'
                          ' your final grade')
        return '\n\n'.join(output)



import os
import re
import json
from fractions import Fraction

from ..submission import BrokenSubmissionError
from ..utils import run_process, PIPE, STDOUT, TimeoutExpired
from ..grades import PartGrade
from . import GraderInterface, Part

"""
Grade a homework with JUnit tests using the log format created by
Patrick (thanks Patrick!)
"""


class JUnitJSONTest(Part):
    __slots__ = ('cls', 'name')

    def __init__(self, test):
        self.test = test
        self.cls, self.name = test.rsplit('.', 1)

    def description(self):
        return '{}.{}'.format(self.cls.rsplit('.', 1)[-1], self.name)


class JUnitJSONGrader(GraderInterface):
    """
    Run a grader jar file and parse the gradelog.json log file the
    creates. This is something Mr. Patrick made just for Zucchini, and
    graders will need to be written to produce this log file.
    """

    DEFAULT_TIMEOUT = 5
    CLASS_REGEX = re.compile(r'\[engine:.*?\]/\[class:(?P<cls>.*?)\]')

    def __init__(self, grader_jar, timeout=None):
        self.grader_jar = grader_jar

        if timeout is None:
            self.timeout = self.DEFAULT_TIMEOUT
        else:
            self.timeout = timeout

    def list_prerequisites(self):
        return ['openjdk-8-jre-headless']

    def part_from_config_dict(self, config_dict):
        return JUnitJSONTest.from_config_dict(config_dict)

    def grade(self, submission, path, parts):
        try:
            process = run_process(['java', '-jar', self.grader_jar],
                                  cwd=path,
                                  timeout=self.timeout,
                                  stdout=PIPE,
                                  stderr=STDOUT)
        except TimeoutExpired:
            raise BrokenSubmissionError('timeout of {} seconds expired for '
                                        'grader'.format(self.timeout))

        if process.returncode != 0:
            raise BrokenSubmissionError(
                'grader command exited with nonzero exit code {}'
                .format(process.returncode),
                verbose=process.stdout.decode() if process.stdout else None)

        gradelog_path = os.path.join(path, 'gradelog.json')

        if not os.path.exists(gradelog_path):
            raise BrokenSubmissionError('gradelog.json does not exist')

        with open(gradelog_path) as gradelog_file:
            gradelog = json.load(gradelog_file)

        results = {}

        for result in gradelog:
            cls = self.CLASS_REGEX.match(result['parentId']).group('cls')
            name = result['displayName']
            score = Fraction(result['status'] == 'PASS')
            log = result['failDescription']
            results[(cls, name)] = PartGrade(score=score, log=log)

        return [results[(part.cls, part.name)] for part in parts]



import os
import re
import shlex
import fnmatch
from fractions import Fraction
import xml.etree.ElementTree
from os import listdir
from os.path import isfile, join

from ..submission import BrokenSubmissionError
from ..utils import run_process, PIPE, STDOUT, TimeoutExpired
from ..grades import PartGrade
from . import GraderInterface, Part, InvalidGraderConfigError

"""
Grade a homework with JUnit tests using Apache Ant formatted XML files
"""


class JUnitXMLTest(Part):
    __slots__ = ('cls', 'name')

    def __init__(self, test):
        self.test = test
        self.cls, self.name = test.rsplit('.', 1)

    def description(self):
        return '{}.{}'.format(self.cls.rsplit('.', 1)[-1], self.name)


class JUnitXMLGrader(GraderInterface):
    """
    Run a gradle test task and parses the resulting XML file, which lists
    successful and failing tests.
    """

    DEFAULT_TIMEOUT = 20
    DEFAULT_RESULT_DIR = 'build/test-results/test'
    CLASS_REGEX = re.compile(r'\[engine:.*?\]/\[class:(?P<cls>.*?)\]')
    DEFAULT_RESULT_MATCHER = 'TEST-*.xml'

    def __init__(self, gradle_task, windows_gradle_exec=None,
                 posix_gradle_exec=None, timeout=None, xml_result_dir=None,
                 result_matcher=None):
        if os.name == 'nt':
            if windows_gradle_exec is None:
                raise InvalidGraderConfigError(
                    'running on windows without windows-gradle-exec set!')
            gradle_exec = windows_gradle_exec
        else:
            # Assume that if it's not windows, it's POSIX
            if posix_gradle_exec is None:
                raise InvalidGraderConfigError(
                    'running on posix without posix-gradle-exec set!')
            gradle_exec = posix_gradle_exec

        self.gradle_cmd = shlex.split(gradle_exec) + [gradle_task]
        self.timeout = self.DEFAULT_TIMEOUT if timeout is None else timeout
        self.xml_result_dir = self.DEFAULT_RESULT_DIR \
            if xml_result_dir is None else xml_result_dir
        self.result_matcher = self.DEFAULT_RESULT_MATCHER \
            if result_matcher is None else result_matcher
        self.use_shell = os.name == 'nt'

    def list_prerequisites(self):
        return ['openjdk-8-jre-headless']

    def part_from_config_dict(self, config_dict):
        return JUnitXMLTest.from_config_dict(config_dict)

    def grade(self, submission, path, parts):
        gradle_cmd = self.gradle_cmd[:]

        for part in parts:
            gradle_cmd += ['--tests', part.test]

        try:
            process = run_process(gradle_cmd, cwd=path, timeout=self.timeout,
                                  stdout=PIPE, stderr=STDOUT,
                                  shell=self.use_shell, input='')
        except TimeoutExpired:
            raise BrokenSubmissionError('timeout of {} seconds expired for '
                                        'grader'.format(self.timeout))

        # DO NOT check gradle command return code because a nonzero exit code
        # is returned if tests fail, obviously this is expected
        # if process.returncode != 0:
        #     raise BrokenSubmissionError(
        #         'grader command exited with nonzero exit code {}'
        #         .format(process.returncode),
        #         verbose=process.stdout.decode() if process.stdout else None)

        test_result_path = os.path.join(path, self.xml_result_dir)

        if not os.path.exists(test_result_path):
            # XXX Is BrokenSubmissionError the right exception to use here?
            raise BrokenSubmissionError('could not find test results dir',
                                        verbose=process.stdout.decode()
                                        if process.stdout else None)

        # goes into xml_result_dir (relative path) and looks for files matching
        # result_matcher
        files = [f for f in listdir(test_result_path)
                 if isfile(join(test_result_path, f))
                 and fnmatch.fnmatch(f, self.result_matcher)]

        # processes each XML result file and sends results back
        results = {}
        for f in files:
            test_xml = os.path.join(test_result_path, f)
            e = xml.etree.ElementTree.parse(test_xml).getroot()
            for testcase in e.findall('testcase'):
                failures = testcase.findall('failure')
                cls = testcase.get('classname')
                name = testcase.get('name')
                score = Fraction(1 if not failures else 0)
                log = failures[0].get('message') if failures else ""
                results[(cls, name)] = PartGrade(score=score, log=log)

        return [results[(part.cls, part.name)] for part in parts]



import collections
from fractions import Fraction

import click

from ..grades import PartGrade
from . import GraderInterface, InvalidGraderConfigError, Part


class Prompt(Part):
    def __init__(self, text, answer_type, answer_range=None):
        self.text = text
        self.answer_type = None

        if answer_type == bool or answer_type == "bool":
            self.answer_type = bool
            self.range_low = 0
            self.range_high = 1
        elif answer_type == int or answer_type == "int":
            self.answer_type = int
        else:
            raise InvalidGraderConfigError("Invalid answer_type: %s. Only bool"
                                           " and int are supported." %
                                           answer_type)

        if self.answer_type == int:  # In this case we also need a range
            if answer_range is None:
                raise InvalidGraderConfigError("An answer_range needs to be "
                                               "specified for int prompts.")

            if not isinstance(answer_range, collections.Iterable):
                raise InvalidGraderConfigError("The answer_range needs to be a"
                                               " list of two elements: start"
                                               " and end.")

            self.answer_range = tuple(answer_range)

            if len(self.answer_range) != 2:
                raise InvalidGraderConfigError("Prompt score answer_range "
                                               "should consist only of two "
                                               "elements.")

            for element in self.answer_range:
                if type(element) != int:
                    raise InvalidGraderConfigError(
                        "Types of the answer_range elements need to be int.")

            self.range_low, self.range_high = self.answer_range
            self.answer_type = click.IntRange(self.range_low, self.range_high)

    def description(self):
        return self.text

    def grade(self):
        """Prompt the user and return a SubcomponentGrade for this prompt"""

        response = click.prompt(text=self.text, type=self.answer_type)
        score = Fraction(response - self.range_low,
                         self.range_high - self.range_low)
        return PartGrade(score=score, log='response: {}'.format(response))


class PromptGrader(GraderInterface):
    def is_interactive(self):
        return True

    def part_from_config_dict(self, config_dict):
        return Prompt.from_config_dict(config_dict)

    def grade(self, submission, path, parts):
        # The submission is not relevant here, so don't use it
        return [prompt.grade() for prompt in parts]



from fractions import Fraction

from .utils import ConfigDictMixin, Record

"""Store grades for components and parts."""


class AssignmentComponentGrade(ConfigDictMixin):
    """Hold the score for an assignment component."""

    def __init__(self, part_grades=None, error=None, error_verbose=None):
        self.part_grades = part_grades
        self.error = error
        self.error_verbose = error_verbose

        if (self.part_grades is None) == (self.error is None):
            raise ValueError('need to specify either part-grades or error in '
                             'an AssignmentComponentGrade, but not both')

    def __repr__(self):
        return '<AssignmentComponentGrade part_grades={}>' \
               .format(self.part_grades)

    @classmethod
    def from_config_dict(cls, dict_):
        grade = super(AssignmentComponentGrade, cls).from_config_dict(dict_)
        if grade.part_grades:
            grade.part_grades = [PartGrade.from_config_dict(g)
                                 for g in grade.part_grades]
        return grade

    def to_config_dict(self, *args):
        dict_ = super(AssignmentComponentGrade, self).to_config_dict(*args)
        if dict_.get('part-grades', None):
            dict_['part-grades'] = [g.to_config_dict()
                                    for g in dict_['part-grades']]
        return dict_

    def is_broken(self):
        """
        Return True if and only if this submission was 'broken'; that
        is, processing it produced an unrecoverable error such as a
        missing file or noncompiling code.
        """
        return self.error is not None

    def calculate_grade(self, points, name, total_part_weight,
                        component_parts):
        """
        Using the list of ComponentPart instances provided (which
        contain the weight of components) and the part grades held in
        this instance, calculate the CalculatedComponentGrade tree for
        this grade.
        """

        grade = CalculatedComponentGrade(name=name,
                                         points_delta=Fraction(0),
                                         points_got=Fraction(0),
                                         points_possible=Fraction(0),
                                         grade=Fraction(1),
                                         error=None,
                                         error_verbose=None,
                                         parts=[])

        grade.grade = Fraction(0)

        if self.is_broken():
            grade.points_got = Fraction(0)
            grade.error = self.error
            grade.error_verbose = self.error_verbose
        else:
            for part, part_grade in zip(component_parts, self.part_grades):
                calc_part_grade = part.calculate_grade(
                    points, total_part_weight, part_grade)
                grade.parts.append(calc_part_grade)
                grade.points_got += calc_part_grade.points_got
                grade.grade += part_grade.score * Fraction(part.weight,
                                                           total_part_weight)

        grade.points_possible = points
        grade.points_delta = grade.points_got - grade.points_possible

        return grade


class PartGrade(ConfigDictMixin):
    """
    Hold the results of grading one part.

    score is the percentage passed as a Fraction instance, deductions is
    a list of deduction ids, and log is a string containing verbose logs
    for this part.
    """

    __slots__ = ('score', 'deductions', 'log')

    def __init__(self, score, deductions=None, log=None):
        self.score = Fraction(score)
        self.deductions = deductions
        self.log = log

    def __repr__(self):
        return '<PartGrade score={}, deductions={}, log={}>' \
               .format(self.score, self.deductions, self.log)

    def to_config_dict(self, *exclude):
        result = super(PartGrade, self).to_config_dict(exclude)
        # Convert Fraction instance to a string
        result['score'] = str(result['score'])
        return result

    @classmethod
    def from_config_dict(cls, config_dict):
        part_grade = super(PartGrade, cls).from_config_dict(config_dict)
        # Convert string to Fraction instance
        part_grade.score = Fraction(part_grade.score)
        return part_grade

    def calculate_grade(self, points, part, partial_credit):
        points_got = self.score * points
        if not partial_credit and points_got < points:
            points_got = Fraction(0)

        return CalculatedPartGrade(name=part.description(),
                                   points_delta=points_got - points,
                                   points_got=points_got,
                                   points_possible=points,
                                   grade=self.score,
                                   deductions=self.deductions,
                                   log=self.log)


class CalculatedGrade(Record):
    """
    Hold the results of grading an assignment. Any numbers are a
    Fraction instance representing actual over possible.
    """
    __slots__ = ['name', 'grade', 'raw_grade', 'penalties', 'components']


class CalculatedPenalty(Record):
    """
    Hold the result of applying (or not applying) a penalty. Any numbers
    are a Fraction instance representing actual over possible.
    """
    __slots__ = ['name', 'points_delta']


class CalculatedComponentGrade(Record):
    """
    Hold the result of grading an assignment component. If error is not
    None, it is an error message string explaining why the submission is
    broken. Any numbers are a Fraction instance representing actual over
    possible.
    """
    __slots__ = ['name', 'points_delta', 'points_got', 'points_possible',
                 'grade', 'error', 'error_verbose', 'parts']


class CalculatedPartGrade(Record):
    """
    Hold the result of grading a single part (test) of an assignment
    component. Any numbers are a Fraction instance representing actual
    over possible.
    """
    __slots__ = ['name', 'points_delta', 'points_got', 'points_possible',
                 'grade', 'deductions', 'log']



"""Provide submission filtering logic."""


class FilterableSubmission(object):
    """Provide an abstract way of getting information about a submission"""

    def __init__(self, base_obj):
        raise NotImplementedError

    def student_name(self):
        raise NotImplementedError

    def is_broken(self):
        raise NotImplementedError


class FilterableMetaSubmission(FilterableSubmission):
    """Filters a Canvas API submission response"""

    __slots__ = ('submission')

    def __init__(self, submission):
        self.submission = submission

    def student_name(self):
        return self.submission.student_name

    def is_broken(self):
        return self.submission.is_broken()


class FilterableCanvasSubmission(FilterableSubmission):
    """Filters a Canvas API submission response"""

    __slots__ = ('canvas_submission')

    def __init__(self, canvas_submission):
        self.canvas_submission = canvas_submission

    def student_name(self):
        return self.canvas_submission.user.sortable_name


class FilterCondition(object):
    """
    Base class for a submission filter condition.

    At least one sufficient condition matching will cause a filter match
    provided all necessary conditions match. If the filter doesn't
    really make sense, not applicable will exclude it entirely.
    """
    __slots__ = ()

    SUFFICIENT, NECESSARY, NOT_APPLICABLE = range(3)

    def type(self):
        """
        Return whether this condition is sufficient or necessary.
        """
        pass

    def accepts(self, submission):
        """Return True if this submission matches the condition, """
        pass


class FilterBrokenCondition(object):
    """Filter for broken submissions"""
    __slots__ = ('is_broken')

    def __init__(self, is_broken):
        self.is_broken = is_broken

    def type(self):
        return FilterCondition.NECESSARY

    def accepts(self, submission):
        # XNOR (pronounced by Conte as "snore")
        return submission.is_broken() == self.is_broken


class FilterStudentCondition(object):
    """Filter by student name"""
    __slots__ = ('student_name')

    def __init__(self, student_name):
        self.student_name = student_name

    def type(self):
        return FilterCondition.SUFFICIENT

    def accepts(self, submission):
        return submission.student_name() == self.student_name


class FilterNotStudentCondition(object):
    """Exclude by student name"""
    __slots__ = ('student_name')

    def __init__(self, student_name):
        self.student_name = student_name

    def type(self):
        return FilterCondition.NECESSARY

    def accepts(self, submission):
        return submission.student_name() != self.student_name


class FilterBuilder(object):
    """
    Filter assignment submissions. Considers the logical OR of all
    conditions added. If no conditions are added, will match all
    submissions.
    """

    def __init__(self, filterable_class):
        self.filterable_class = filterable_class
        self.conditions = []

    @classmethod
    def new_canvas(cls):
        """Return a FilterBuilder ready to handle Canvas API responses"""
        return cls(FilterableCanvasSubmission)

    @classmethod
    def new_meta(cls):
        """Return a FilterBuilder ready to handle submission metadata"""
        return cls(FilterableMetaSubmission)

    def __call__(self, submission):
        # Convert this lower-level object into something we can pass to
        # the higher-level filters
        filterable = self.filterable_class(submission)

        necessary_met = True
        sufficient_met = False
        # Still want to accept a submission if we have no sufficient
        # conditions
        found_sufficient_condition = False
        for cond in self.conditions:
            try:
                result = cond.accepts(filterable)
            except NotImplementedError:
                # Some filters aren't implemented for some submission
                # types. For example, it doesn't make much sense for a
                # Canvas API submission to be 'broken'
                continue

            if cond.type() == FilterCondition.NECESSARY:
                necessary_met = necessary_met and result
            elif cond.type() == FilterCondition.SUFFICIENT:
                found_sufficient_condition = True
                sufficient_met = sufficient_met or result

        return necessary_met \
            and (not found_sufficient_condition or sufficient_met)

    def add_student_name(self, student_name):
        self.conditions.append(FilterStudentCondition(student_name))

    def add_not_student_name(self, student_name):
        self.conditions.append(FilterNotStudentCondition(student_name))

    def add_broken(self, is_broken):
        self.conditions.append(FilterBrokenCondition(is_broken))



import os

import click

from . import PromptGrader, InvalidGraderConfigError
from ..utils import sanitize_path


class OpenFileGrader(PromptGrader):
    def __init__(self, file_name):
        # Set up the Prompt Grader first
        super(OpenFileGrader, self).__init__()

        if not isinstance(file_name, str) or not file_name:
            raise InvalidGraderConfigError(
                "A file_name needs to be specified.")
        else:
            self.file_name = sanitize_path(file_name)

    def grade(self, submission, path, parts):
        # Let's get the file from the submission
        file_to_launch = os.path.join(path, self.file_name)
        click.launch(file_to_launch)

        return super(OpenFileGrader, self).grade(submission=submission,
                                                 path=path, parts=parts)



import os
import json

from .grades import AssignmentComponentGrade
from .constants import SUBMISSION_META_FILE, SUBMISSION_FILES_DIRECTORY
from .utils import ConfigDictMixin, datetime_to_string, copy_globs, \
                   FileNotFoundError


class BrokenSubmissionError(Exception):
    def __init__(self, message, verbose=None):
        super(BrokenSubmissionError, self).__init__(message)
        self.message = message
        self.verbose = verbose


class Submission(ConfigDictMixin):
    def __init__(self, student_name, assignment, metadata_path, files_path,
                 graded, id=None, seconds_late=None, error=None,
                 component_grades=None):
        self.student_name = student_name
        self.assignment = assignment
        self.metadata_path = metadata_path
        self.files_path = files_path
        self.graded = graded
        self.id = id
        self.seconds_late = seconds_late

        if component_grades is not None and error is not None:
            raise ValueError('either specify component_grades or error, '
                             'but not both')

        self.error = error

        if component_grades is None:
            self.component_grades = None
        else:
            self.component_grades = [
                AssignmentComponentGrade.from_config_dict(grade_dict)
                for grade_dict in component_grades]

        # TODO: Validate the path - make sure everything that's needed for the
        # assignment is available in the path

    @staticmethod
    def get_paths(path):
        """
        Calculate the metadata and files paths for a standard submission
        from `zucc load'.
        """
        metadata_path = os.path.join(path, SUBMISSION_META_FILE)
        files_path = os.path.join(path, SUBMISSION_FILES_DIRECTORY)

        return metadata_path, files_path

    @classmethod
    def load_from_empty_dir(cls, assignment, path, **kwargs):
        """
        Load a Submission instance from an uninitialized submission directory.
        """

        metadata_path, files_path = cls.get_paths(path)
        return cls(assignment=assignment, metadata_path=metadata_path,
                   files_path=files_path, **kwargs)

    @classmethod
    def load_from_dir(cls, assignment, path):
        """Load a Submission instance from a submission directory."""

        metadata_path, files_path = cls.get_paths(path)

        with open(metadata_path) as meta_file:
            meta_json = json.load(meta_file)

        return cls.from_config_dict(meta_json, assignment=assignment,
                                    metadata_path=metadata_path,
                                    files_path=files_path)

    @classmethod
    def load_from_raw_files(cls, assignment, files_path):
        return Submission(student_name='', assignment=assignment,
                          metadata_path=None, files_path=files_path,
                          graded=False)

    @classmethod
    def load_from_component_grades_json(cls, assignment, component_grades_fp,
                                        seconds_late=None):
        component_grades = json.load(component_grades_fp)

        return Submission(student_name='', assignment=assignment,
                          metadata_path=None, files_path=None, graded=True,
                          seconds_late=seconds_late,
                          component_grades=component_grades)

    def _meta_json(self):
        """Return a json representation of this instance"""

        meta = self.to_config_dict('assignment', 'metadata_path', 'files_path')
        if 'submission-time' in meta:
            meta['submission-time'] = datetime_to_string(self.submission_time)

        return meta

    def _write_meta_json(self):
        """Json-ify this instance and write to the metadata json file"""

        meta = self._meta_json()

        with open(self.metadata_path, 'w') as meta_file:
            json.dump(meta, meta_file, sort_keys=True, indent=2,
                      separators=(',', ': '))

    def initialize_metadata(self):
        """Create initial meta.json"""

        self._write_meta_json()

    def is_broken(self):
        """
        Return true if this assignment has been graded but at least one
        component grade flagged a broken submisison.
        """

        return self.error is not None or self.component_grades is not None \
            and any(component.is_broken()
                    for component in self.component_grades)

    # XXX Support copying directories
    def copy_files(self, files, path, allow_fail=False):
        try:
            copy_globs(files, self.files_path, path)
        except FileNotFoundError as err:
            if not allow_fail:
                raise BrokenSubmissionError(str(err))

    def write_grade(self, component_grades):  # (Dict[object, object]) -> None
        """
        Set the component grades to `component_grades' and write the
        new submission metadata to the metadata file.
        """

        self.graded = True
        self.error = None
        self.component_grades = component_grades
        self._write_meta_json()
